from __future__ import annotations

from _typeshed import (
    StrOrBytesPath,
    SupportsWrite,
    Incomplete,
    BytesPath,
    StrPath,
    Unused
)

from collections.abc import (
    Callable,
    Iterable,
    Mapping
)

from typing import (
    overload,
    ClassVar,
    Literal,
    TypeVar,
    Any,
    IO
)

from typing_extensions import TypeAlias
from abc import abstractmethod

_OptionsList: TypeAlias = list[tuple[str, str | None, str, int] | tuple[str, str | None, str]]
_CommandT = TypeVar("_CommandT", bound=Command)

_StrPathT = TypeVar("_StrPathT", bound=StrPath)
_BytesPathT = TypeVar("_BytesPathT", bound=BytesPath)

class Command:
    distribution: Distribution
    # Any to work around variance issues
    sub_commands: ClassVar[list[tuple[str, Callable[[Any], bool] | None]]]
    def __init__(self, dist: Distribution) -> None: ...
    def ensure_finalized(self) -> None: ...
    @abstractmethod
    def initialize_options(self) -> None: ...
    @abstractmethod
    def finalize_options(self) -> None: ...
    @abstractmethod
    def run(self) -> None: ...
    def announce(self, msg: str, level: int = ...) -> None: ...
    def debug_print(self, msg: str) -> None: ...
    def ensure_string(self, option: str, default: str | None = ...) -> None: ...
    def ensure_string_list(self, option: str | list[str]) -> None: ...
    def ensure_filename(self, option: str) -> None: ...
    def ensure_dirname(self, option: str) -> None: ...
    def get_command_name(self) -> str: ...
    def set_undefined_options(self, src_cmd: str, *option_pairs: tuple[str, str]) -> None: ...
    def get_finalized_command(self, command: str, create: bool | Literal[0, 1] = 1) -> Command: ...
    def reinitialize_command(self, command: Command | str, reinit_subcommands: bool | Literal[0, 1] = 0) -> Command: ...
    def run_command(self, command: str) -> None: ...
    def get_sub_commands(self) -> list[str]: ...
    def warn(self, msg: str) -> None: ...
    def execute(
        self, func: Callable[..., object], args: Iterable[Incomplete], msg: str | None = ..., level: int = ...
    ) -> None: ...
    def mkpath(self, name: str, mode: int = ...) -> None: ...
    @overload
    def copy_file(
        self,
        infile: StrPath,
        outfile: _StrPathT,
        preserve_mode: bool | Literal[0, 1] = 1,
        preserve_times: bool | Literal[0, 1] = 1,
        link: str | None = None,
        level: Unused = 1,
    ) -> tuple[_StrPathT | str, bool]: ...
    @overload
    def copy_file(
        self,
        infile: BytesPath,
        outfile: _BytesPathT,
        preserve_mode: bool | Literal[0, 1] = 1,
        preserve_times: bool | Literal[0, 1] = 1,
        link: str | None = None,
        level: Unused = 1,
    ) -> tuple[_BytesPathT | bytes, bool]: ...
    def copy_tree(
        self,
        infile: StrPath,
        outfile: str,
        preserve_mode: bool | Literal[0, 1] = 1,
        preserve_times: bool | Literal[0, 1] = 1,
        preserve_symlinks: bool | Literal[0, 1] = 0,
        level: Unused = 1,
    ) -> list[str]: ...
    @overload
    def move_file(self, src: StrPath, dst: _StrPathT, level: Unused = 1) -> _StrPathT | str: ...
    @overload
    def move_file(self, src: BytesPath, dst: _BytesPathT, level: Unused = 1) -> _BytesPathT | bytes: ...
    def spawn(self, cmd: Iterable[str], search_path: bool | Literal[0, 1] = 1, level: Unused = 1) -> None: ...
    @overload
    def make_archive(
        self,
        base_name: str,
        format: str,
        root_dir: StrOrBytesPath | None = None,
        base_dir: str | None = None,
        owner: str | None = None,
        group: str | None = None,
    ) -> str: ...
    @overload
    def make_archive(
        self,
        base_name: StrPath,
        format: str,
        root_dir: StrOrBytesPath,
        base_dir: str | None = None,
        owner: str | None = None,
        group: str | None = None,
    ) -> str: ...
    def make_file(
        self,
        infiles: str | list[str] | tuple[str, ...],
        outfile: StrOrBytesPath,
        func: Callable[..., object],
        args: list[Incomplete],
        exec_msg: str | None = None,
        skip_msg: str | None = None,
        level: Unused = 1,
    ) -> None: ...

class DistributionMetadata:
    def __init__(self, path: StrOrBytesPath | None = None) -> None: ...
    name: str | None
    version: str | None
    author: str | None
    author_email: str | None
    maintainer: str | None
    maintainer_email: str | None
    url: str | None
    license: str | None
    description: str | None
    long_description: str | None
    keywords: str | list[str] | None
    platforms: str | list[str] | None
    classifiers: str | list[str] | None
    download_url: str | None
    provides: list[str] | None
    requires: list[str] | None
    obsoletes: list[str] | None
    def read_pkg_file(self, file: IO[str]) -> None: ...
    def write_pkg_info(self, base_dir: StrPath) -> None: ...
    def write_pkg_file(self, file: SupportsWrite[str]) -> None: ...
    def get_name(self) -> str: ...
    def get_version(self) -> str: ...
    def get_fullname(self) -> str: ...
    def get_author(self) -> str: ...
    def get_author_email(self) -> str: ...
    def get_maintainer(self) -> str: ...
    def get_maintainer_email(self) -> str: ...
    def get_contact(self) -> str: ...
    def get_contact_email(self) -> str: ...
    def get_url(self) -> str: ...
    def get_license(self) -> str: ...
    def get_licence(self) -> str: ...
    def get_description(self) -> str: ...
    def get_long_description(self) -> str: ...
    def get_keywords(self) -> str | list[str]: ...
    def get_platforms(self) -> str | list[str]: ...
    def get_classifiers(self) -> str | list[str]: ...
    def get_download_url(self) -> str: ...
    def get_requires(self) -> list[str]: ...
    def set_requires(self, value: Iterable[str]) -> None: ...
    def get_provides(self) -> list[str]: ...
    def set_provides(self, value: Iterable[str]) -> None: ...
    def get_obsoletes(self) -> list[str]: ...
    def set_obsoletes(self, value: Iterable[str]) -> None: ...

class Distribution:
    cmdclass: dict[str, type[Command]]
    metadata: DistributionMetadata
    def __init__(self, attrs: Mapping[str, Any] | None = None) -> None: ...
    def get_option_dict(self, command: str) -> dict[str, tuple[str, str]]: ...
    def parse_config_files(self, filenames: Iterable[str] | None = None) -> None: ...
    @overload
    def get_command_obj(self, command: str, create: Literal[1, True] = 1) -> Command: ...
    @overload
    def get_command_obj(self, command: str, create: Literal[0, False]) -> Command | None: ...
    global_options: ClassVar[_OptionsList]
    common_usage: ClassVar[str]
    display_options: ClassVar[_OptionsList]
    display_option_names: ClassVar[list[str]]
    negative_opt: ClassVar[dict[str, str]]
    verbose: int
    dry_run: int
    help: int
    command_packages: list[str] | None
    script_name: str | None
    script_args: list[str] | None
    command_options: dict[str, dict[str, tuple[str, str]]]
    dist_files: list[tuple[str, str, str]]
    packages: Incomplete
    package_data: dict[str, list[str]]
    package_dir: Incomplete
    py_modules: Incomplete
    libraries: Incomplete
    headers: Incomplete
    ext_modules: Incomplete
    ext_package: Incomplete
    include_dirs: Incomplete
    extra_path: Incomplete
    scripts: Incomplete
    data_files: Incomplete
    password: str
    command_obj: dict[str, Command]
    have_run: dict[str, bool]
    want_user_cfg: bool
    def dump_option_dicts(
        self, header: Incomplete | None = None, commands: Incomplete | None = None, indent: str = ""
    ) -> None: ...
    def find_config_files(self): ...
    commands: Incomplete
    def parse_command_line(self): ...
    def finalize_options(self) -> None: ...
    def handle_display_options(self, option_order): ...
    def print_command_list(self, commands, header, max_length) -> None: ...
    def print_commands(self) -> None: ...
    def get_command_list(self): ...
    def get_command_packages(self): ...
    def get_command_class(self, command: str) -> type[Command]: ...
    @overload
    def reinitialize_command(self, command: str, reinit_subcommands: bool = False) -> Command: ...
    @overload
    def reinitialize_command(self, command: _CommandT, reinit_subcommands: bool = False) -> _CommandT: ...
    def announce(self, msg, level: int = ...) -> None: ...
    def run_commands(self) -> None: ...
    def run_command(self, command: str) -> None: ...
    def has_pure_modules(self) -> bool: ...
    def has_ext_modules(self) -> bool: ...
    def has_c_libraries(self) -> bool: ...
    def has_modules(self) -> bool: ...
    def has_headers(self) -> bool: ...
    def has_scripts(self) -> bool: ...
    def has_data_files(self) -> bool: ...
    def is_pure(self) -> bool: ...

    # Getter methods generated in __init__
    def get_name(self) -> str: ...
    def get_version(self) -> str: ...
    def get_fullname(self) -> str: ...
    def get_author(self) -> str: ...
    def get_author_email(self) -> str: ...
    def get_maintainer(self) -> str: ...
    def get_maintainer_email(self) -> str: ...
    def get_contact(self) -> str: ...
    def get_contact_email(self) -> str: ...
    def get_url(self) -> str: ...
    def get_license(self) -> str: ...
    def get_licence(self) -> str: ...
    def get_description(self) -> str: ...
    def get_long_description(self) -> str: ...
    def get_keywords(self) -> str | list[str]: ...
    def get_platforms(self) -> str | list[str]: ...
    def get_classifiers(self) -> str | list[str]: ...
    def get_download_url(self) -> str: ...
    def get_requires(self) -> list[str]: ...
    def get_provides(self) -> list[str]: ...
    def get_obsoletes(self) -> list[str]: ...

class Extension:
    name: str
    sources: list[str]
    include_dirs: list[str]
    define_macros: list[tuple[str, str | None]]
    undef_macros: list[str]
    library_dirs: list[str]
    libraries: list[str]
    runtime_library_dirs: list[str]
    extra_objects: list[str]
    extra_compile_args: list[str]
    extra_link_args: list[str]
    export_symbols: list[str]
    swig_opts: list[str]
    depends: list[str]
    language: str | None
    optional: bool | None
    def __init__(
        self,
        name: str,
        sources: list[str],
        include_dirs: list[str] | None = None,
        define_macros: list[tuple[str, str | None]] | None = None,
        undef_macros: list[str] | None = None,
        library_dirs: list[str] | None = None,
        libraries: list[str] | None = None,
        runtime_library_dirs: list[str] | None = None,
        extra_objects: list[str] | None = None,
        extra_compile_args: list[str] | None = None,
        extra_link_args: list[str] | None = None,
        export_symbols: list[str] | None = None,
        swig_opts: list[str] | None = None,
        depends: list[str] | None = None,
        language: str | None = None,
        optional: bool | None = None,
    ) -> None: ...


from typehintwhl.exception import ConfigurationError, SetupError

from typehintwhl.common import (
    process_exclude_list, 
    remove_string_quotes, 
    remove_string_affix,
    get_type_name
)

from pathlib import Path
from types import NoneType

import configparser
import tomllib

class Config:
    def __init__(self, package, stub, current=None, exclude=None):
        self.stub = stub
        self.package = package
        self.current = current
        self.exclude = exclude
    
    def __repr__(self):
        return f"Config(package={self.package}, stub={self.stub}, current={self.current}, exclude={self.exclude})"

class ConfigLoader:
    class _Tracer:
        def __init__(self):
            self.stub = None
            self.package = None
            self.exclude = []
        
        def set(self, name, value):
            # Don't override already set values
            if getattr(self, name) is None:
                setattr(self, name, Path(value).absolute() if isinstance(value, str) else value)
        
        def integrate(self, config):
            if config is None or self.check():
                return

            self.set("stub", config.stub)
            self.set("package", config.package)
            self.exclude.extend(config.exclude or [])
        
        def check(self):
            return self.package is not None and self.stub is not None
        
        @property
        def missing(self):
            return list([v[0] for v in filter(lambda v: v[1] is None, [("package", self.package,), ("stub", self.stub,)])])

    @staticmethod
    def pyproject(file):
        """Load configuration from a pyproject.toml file."""
        file = file.absolute() # Verify absolute path
        
        with open(file, "rb") as fp:
            data = tomllib.load(fp)
        
        config = data.get("tool", {}).get("typehintwhl", {})

        package, stub, exclude = ConfigLoader.process_config_data(
            file,
            config.get("package"), 
            config.get("stub"), 
            config.get("exclude")
        )

        return Config(package, stub, file, exclude)
    
    @staticmethod
    def setupcfg(file):
        """Load configuration from a setup.cfg file."""
        file = file.absolute() # Verify absolute path

        cfg = configparser.RawConfigParser()
        cfg.read(file.absolute())
        cfg = cfg["typehintwhl"] if cfg.has_section("typehintwhl") else None

        if cfg is None:
            return

        package, stub, exclude = ConfigLoader.process_config_data(
            file,
            remove_string_quotes(cfg.get("package", fallback=None)),
            remove_string_quotes(cfg.get("stub", fallback=None)),
            remove_string_quotes(cfg.get("exclude", fallback=None))
        )

        if package is None and stub is None and not exclude:
            return

        return Config(package, stub, file, exclude)
    
    @staticmethod
    def cli(args):
        """Load configuration from command line arguments."""
        
        package, stub, exclude = ConfigLoader.process_config_data(
            args.current,
            args.package,
            args.stub,
            args.exclude or []
        )

        if package is None and stub is None and not exclude:
            return
        
        return Config(package, stub, exclude)

    @staticmethod
    def loadAllFiles(file):
        file = file.absolute() # Verify absolute path
        pyproject_path = Path(file, "pyproject.toml").absolute()
        setupcfg_path = Path(file, "setup.cfg").absolute()

        tracer = ConfigLoader._Tracer()

        tracer.integrate(ConfigLoader.pyproject(pyproject_path) if pyproject_path.exists() else None)
        tracer.integrate(ConfigLoader.setupcfg(setupcfg_path) if setupcfg_path.exists() else None)

        if tracer.check():
            return Config(tracer.package, tracer.stub, tracer.exclude)
    
    @staticmethod
    def loadAll(file, args):
        # Load Order (cli, pyproject.toml, setup.cfg)
        file = file.absolute() # Verify absolute path
        tracer = ConfigLoader._Tracer()

        tracer.integrate(ConfigLoader.cli(args) if args is not None else None)
        tracer.integrate(ConfigLoader.loadAllFiles(file))

        if tracer.check():
            return Config(tracer.package, tracer.stub, file, tracer.exclude)
        
        msg = "\n".join([
            f"Failed to complete configuration, not enough information.", 
            "\tMissing: ", remove_string_affix(repr(tracer.missing), '[', ']').replace("'", "")
        ])

        raise SetupError(msg)

    @staticmethod
    def process_config_data(root, package, stub, exclude):
        base_err_msg = (f"'{root.name}' file contains invalid 'typehintwhl' tool configuration.\n\t", f"\n\tConfig Path: ({root})",)
        type_err_msg = base_err_msg[0] + "\tExpected option '{option}' to be of type {expects}, instead got '{got}'" + base_err_msg[1]
        exists_err_msg = base_err_msg[0] + "\t{option} option contains path that does not exist.\n\tSpecified Path: ({path})" + base_err_msg[1]
        dir_err_msg = base_err_msg[0] + "\tExpected option '{option}' to be a directory path."

        if not isinstance(package, (str, NoneType,)):
            raise ConfigurationError(type_err_msg.format(option="package", expects="str", got=get_type_name(package)))
        
        if not isinstance(stub, (str, NoneType,)):
            raise ConfigurationError(type_err_msg.format(option="stub", expects="str", got=get_type_name(stub)))

        if package is not None:
            package = Path(package).absolute()

            if not package.exists():
                raise ConfigurationError(exists_err_msg.format(option="package", path=package))

            if not package.is_dir():
                raise ConfigurationError(dir_err_msg.format(option="package"))
        
        if stub is not None:
            stub = Path(stub).absolute()

            if not stub.exists():
                raise ConfigurationError(exists_err_msg.format(option="stub", path=stub))

            if not stub.is_dir():
                raise ConfigurationError(dir_err_msg.format(option="stub"))
        
        exclude = process_exclude_list(stub, exclude, lambda msg: ConfigurationError(base_err_msg[0] + msg + base_err_msg[1]))

        return package, stub, exclude


from typehintwhl.config import ConfigLoader
from typehintwhl.build import Builder
from typehintwhl.log import logger
from typehintwhl import __version__
from pathlib import Path
import argparse

__all__ = ["run"]

DESCRIPTION = f"""
Distributes a directory of stub files within a package directory, \
enabling you to neatly organize your stub files.
"""

VERSION = f"""
%(prog)s {__version__}
Copyright (C) 2024 Isaiah Coroama. All rights reserved.
"""

def prepare_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="typehintwhl",
        description=DESCRIPTION,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    parser.add_argument(
        "--package",
        metavar="DIR",
        help="The directory to target."
    )

    parser.add_argument(
        "--stub",
        metavar="DIR",
        help="The directory that contains the stub files to be distributed."
    )

    parser.add_argument(
        "--current",
        metavar="DIR",
        help="Specify the current directory to search through. (Defaults to the current workinkg directory)"
    )

    parser.add_argument(
        "--exclude",
        metavar="NAME",
        help="A list of comma-seperated stub directories or files to ignore."
    )

    parser.add_argument(
        "target",
        nargs=argparse.OPTIONAL,
        metavar="DIR",
        help="The directory to target. (Same as --package)"
    )

    parser.add_argument(
        "-v", "--version",
        action="version",
        version=VERSION
    )

    return parser

def run():
    parser = prepare_parser()
    args, argv = parser.parse_known_args()

    args.package = args.package or args.target
    args.current = args.current or Path.cwd()
    
    if "-h" in argv or "--help" in argv:
        parser.print_help()
        parser.exit()

    logger.info(f"typehintwhl {__version__}")
    config = ConfigLoader.loadAll(args.current, args)

    builder = Builder(config.package, config.stub, config.current, config.exclude)
    builder.create()


from collections.abc import Iterable
from pathlib import Path
from types import LambdaType

def get_type_name(value):
    return type(value).__name__

def remove_string_affix(text, prefix, suffix):
    return text.removeprefix(prefix).removesuffix(suffix)

def process_exclude_list(stub, exclude, error = None):
    paths = []

    if error is None:
        error = ValueError

    if isinstance(exclude, str):
        exclude = exclude.split(',')
        
    for name in exclude or []:
        path = Path(name)

        if path.is_absolute() and not path.exists():
            msg = f"Exclude list contains stub file or directory that does not exist. ('{name}')"
            if isinstance(error, LambdaType): error(msg)
            raise error(msg)
        
        if path.exists():
            paths.append(path.absolute())
            continue

        if stub is None:
            msg = f"Exclude list contains stub file or directory that does not exist. ('{name}')"
            if isinstance(error, LambdaType): error(msg)
            raise error(msg)

        path = Path(stub, name).absolute()

        if not path.exists():
            msg = f"Exclude list contains stub file or directory that does not exist. ('{name}')"
            if isinstance(error, LambdaType): error(msg)
            raise error(msg)

        paths.append(path.absolute())
    
    return paths

def remove_string_quotes(value):
    helper = lambda _str: remove_string_affix(remove_string_affix(_str, '"', '"'), "'", "'")

    if isinstance(value, str):
        return helper(value)
    elif isinstance(value, Iterable):
        return [helper(v) for v in value if isinstance(v, str)]

    return value


from .common import AbsolutePath, RelativePath
from _typeshed import StrPath
from pathlib import Path
from typing import List, Any

class Builder:
    def __init__(self, 
        package: AbsolutePath, 
        stub: AbsolutePath, 
        current: AbsolutePath | None = None, 
        exclude: List[AbsolutePath] | None = None
    ) -> None: ...
    @property
    def pyproject(self) -> AbsolutePath | None: ...
    @property
    def setupcfg(self) -> AbsolutePath | None: ...
    @staticmethod
    def check_directory_argument(name: str, value: Any) -> None: ...
    @staticmethod
    def remove_build_from_path(path: Path | StrPath, build: Path | StrPath) -> RelativePath: ...
    @staticmethod
    def get_relative_stub_path(stub: Path | StrPath, base: Path | StrPath) -> RelativePath: ...
    def get_build_path_for_stub(self, original_stub: AbsolutePath) -> AbsolutePath: ...
    def get_build_directory(self) -> AbsolutePath: ...    
    def create_build_directory(self, path: AbsolutePath) -> None: ...
    def delete_build_directory(self, path: AbsolutePath) -> None: ...
    def collect_stub_files(self) -> List[AbsolutePath]: ...
    def construct_file_source(self, path: RelativePath) -> AbsolutePath: ...
    def construct_file_dest(self, root: AbsolutePath, filename: RelativePath) -> AbsolutePath: ...
    @staticmethod
    def create_copy(src: AbsolutePath, dst: AbsolutePath, logging: bool = False) -> None: ...
    def create(self) -> StrPath: ...


from typing_extensions import TypeAlias
from collections.abc import Iterable
from pathlib import Path

from typing import List, Any
from types import LambdaType

AbsolutePath: TypeAlias = Path
RelativePath: TypeAlias = Path

def copy_stub_file(src: Path, dst: Path) -> None: ...
def get_type_name(value: Any) -> str: ...
def remove_string_affix(text: str, prefix: str, suffix: str) -> str: ...
def process_exclude_list(
    stub: str | None, 
    exclude: List[str] | str, 
    error: LambdaType | Exception | None = None
) -> List[Path]: ...
def remove_string_quotes(value: Iterable[str] | str) -> List[str] | str: ...


from .common import AbsolutePath
from typing import Tuple, List
import argparse

class Config:
    def __init__(self, 
        package: AbsolutePath | None, 
        stub: AbsolutePath | None, 
        current: AbsolutePath | None = None, 
        exclude: List[AbsolutePath] | None = None
    ) -> None: ...
    def __repr__(self) -> str: ...

class ConfigLoader:
    class _Tracer:
        def __init__(self) -> None: ...
        def set(self, name: str, value: AbsolutePath | None) -> None: ...
        def integrate(self, config: Config | None) -> None: ...
        def check(self) -> bool: ...
        @property
        def missing(self) -> List[str]: ...

    @staticmethod
    def pyproject(file: AbsolutePath) -> Config: ...
    @staticmethod
    def setupcfg(file: AbsolutePath) -> Config | None: ...
    @staticmethod
    def cli(args: argparse.Namespace) -> Config | None: ...
    @staticmethod
    def loadAllFiles(file: AbsolutePath) -> Config | None: ...
    @staticmethod
    def loadAll(file: AbsolutePath, args: argparse.Namespace) -> Config: ...
    @staticmethod
    def process_config_data(
        root: AbsolutePath, 
        package: AbsolutePath | None, 
        stub: AbsolutePath | None, 
        exclude: List[AbsolutePath]
    ) -> Tuple[AbsolutePath | None, AbsolutePath | None, List[AbsolutePath]]: ...


__all__ = ["SetupError", "ConfigurationError"]

class SetupError(Exception):
    """An error occured during setup due to invalid build configurations."""
    pass

class BuildError(Exception):
    """An error that occurs during the building, usually if a configuration has not be set."""
    pass

class ConfigurationError(Exception):
    """Invalid configuration was provided for building."""
    pass


from typehintwhl.exception import BuildError
from typehintwhl.log import logger

from pathlib import Path
import shutil
import os

class Builder:
    def __init__(self, package, stub, current=None, exclude=None):
        self.stub = stub
        self.package = package
        self.current = current or Path.cwd()
        self.exclude = exclude or []

        self._pyproject = None
        self._setupcfg = None
    
    @property
    def pyproject(self):
        if self._pyproject is None:
            path = Path(self.current, "pyproject.toml").absolute()

            if not path.exists():
                return
            
            self._pyproject = path
            return path
        
        if self._pyproject.exists():
            return self._pyproject
    
    @property
    def setupcfg(self):
        if self._setupcfg is None:
            path = Path(self.current, "setup.cfg").absolute()

            if not path.exists():
                return
            
            self._setupcfg = path
            return path
        
        if self._setupcfg.exists():
            return self._setupcfg
    
    @staticmethod
    def check_directory_argument(name, value):
        """Raises a BuildError if the argument's value is None"""
        if value is None:
            raise BuildError(f"The {name} directory has not been specified.")
    
    @staticmethod
    def remove_build_from_path(path, build):
        if isinstance(path, str):
            path = Path(path)
        
        if isinstance(build, str):
            build = Path(build)
        
        return Path(str(path.absolute()).removeprefix(str(build.absolute())))

    @staticmethod
    def get_relative_stub_path(stub, base):
        if isinstance(stub, str):
            stub = Path(stub)
        
        if isinstance(base, str):
            base = Path(base)
        
        return Path(str(stub).removeprefix(str(base)))

    def get_build_path_for_stub(self, original_stub):
        # Remove the starting \\ or / else it will consider it as an absolute path.
        return Path(self.get_build_directory(), str(original_stub).removeprefix(str(self.stub)).lstrip('\\/')).absolute()
    
    def get_build_directory(self):
        return Path(self.current, "build/lib", self.package.name)
    
    def create_build_directory(self, path):
        # Remove the old directory if it exists
        self.delete_build_directory(path)

        logger.info(f"copying package {self.package} -> {path}")
        shutil.copytree(str(self.package), str(path))

        # shutil.copytree copies the stub directory if it's
        # with in the project directory, remove it here
        stub_path = path / self.stub.name
        if stub_path.exists():
            shutil.rmtree(str(stub_path))

    def delete_build_directory(self, path):
        if not path.exists():
            return
        
        try:
            logger.info("removing previous build")
            shutil.rmtree(str(path))
        except OSError:
            logger.error("failed to remove previous build.")
            exit(1)
    
    def collect_stub_files(self):
        logger.info(f"collecting stub files in directory {self.stub}")

        stubs = []
        for root, _, files in os.walk(self.stub.absolute()):
            base = Path(self.package.name, self.stub.name)

            for file in files:
                path = Path(root, file).absolute()

                if path.suffix == ".pyi" and path not in self.exclude:
                    logger.info(f"found stub {path}")
                    root_name = root.removeprefix(str(base))
                    stubs.append(Path(base, root_name, file).absolute())
        
        return stubs
    
    def construct_file_source(self, path):
        return (self.stub / path.with_suffix(".pyi")).absolute()
    
    def construct_file_dest(self, root, filename):
        return (root / filename.with_suffix(".pyi")).absolute()
    
    @staticmethod
    def create_copy(src, dst, logging=False):
        if not src.exists():
            raise FileNotFoundError(f"The system cannot find the path specified ('{src.absolute()}')")
        
        if logging:
            logger.info(f"copying {src} -> {dst}")
        
        fsrc = open(src, "rb")
        fdst = open(dst, "wb")

        shutil.copyfileobj(fsrc, fdst)

        fsrc.close()
        fdst.close()
    
    def create(self):
        Builder.check_directory_argument("package", self.package)
        Builder.check_directory_argument("stub", self.stub)

        stub_files = self.collect_stub_files()
        build_path = self.get_build_directory()

        self.create_build_directory(build_path)

        for root, _, files in os.walk(build_path):
            if Path(root).name == "__pycache__":
                continue

            relative_root_path = Builder.remove_build_from_path(root, build_path)
            relative_stub_path = Builder.get_relative_stub_path(self.stub, self.package)
            
            if relative_root_path == relative_stub_path:
                continue

            for filename in files:
                filename = Path(filename)

                # Skip non-source files
                if filename.suffix != ".py":
                    continue
                
                src = self.construct_file_source(filename)
                dst = self.construct_file_dest(root, filename)

                if src in stub_files:
                    stub_files.remove(src)
                    Builder.create_copy(src, dst, logging=True)
        
        for full_stub_path in stub_files:
            stub_build_path = self.get_build_path_for_stub(full_stub_path)
            logger.warn(f"warning: '{full_stub_path}' has no matching source file,\n\tcopying to '{stub_build_path}'")
            Builder.create_copy(full_stub_path, stub_build_path)
        
        logger.info("finished building library with stub files.")
        
        return str(build_path)


# This is a copy of the typeshed stub files for setuptools
# https://github.com/python/typeshed/tree/main/stubs/setuptools/setuptools
from collections.abc import (
    MutableMapping,
    Sequence,
    Iterable, 
    Iterator, 
    Mapping
)

from .distutils import (
    Distribution as _Distribution,
    Extension as _Extension,
    Command as _Command
)

from typing import Literal, Any
from abc import abstractmethod

class Extension(_Extension):
    py_limited_api: bool
    def __init__(
        self,
        name: str,
        sources: list[str],
        include_dirs: list[str] | None = None,
        define_macros: list[tuple[str, str | None]] | None = None,
        undef_macros: list[str] | None = None,
        library_dirs: list[str] | None = None,
        libraries: list[str] | None = None,
        runtime_library_dirs: list[str] | None = None,
        extra_objects: list[str] | None = None,
        extra_compile_args: list[str] | None = None,
        extra_link_args: list[str] | None = None,
        export_symbols: list[str] | None = None,
        swig_opts: list[str] | None = None,
        depends: list[str] | None = None,
        language: str | None = None,
        optional: bool | None = None,
        *,
        py_limited_api: bool = False,
    ) -> None: ...

class Distribution(_Distribution):
    def patch_missing_pkg_info(self, attrs: Mapping[str, Any]) -> None: ...
    src_root: str | None
    dependency_links: list[str]
    setup_requires: list[str]
    def __init__(self, attrs: MutableMapping[str, Any] | None = None) -> None: ...
    def warn_dash_deprecation(self, opt: str, section: str) -> str: ...
    def make_option_lowercase(self, opt: str, section: str) -> str: ...
    def parse_config_files(self, filenames: Iterable[str] | None = None, ignore_option_errors: bool = False) -> None: ...
    def fetch_build_eggs(self, requires: str | Iterable[str]): ...
    def get_egg_cache_dir(self) -> str: ...
    def fetch_build_egg(self, req): ...
    def get_command_class(self, command: str) -> type[Command]: ...
    def include(self, **attrs) -> None: ...
    def exclude_package(self, package: str) -> None: ...
    def has_contents_for(self, package: str) -> bool | None: ...
    def exclude(self, **attrs) -> None: ...
    def get_cmdline_options(self) -> dict[str, dict[str, str | None]]: ...
    def iter_distribution_names(self) -> Iterator[str]: ...
    def handle_display_options(self, option_order): ...

def setup(
    *,
    name: str = ...,
    version: str = ...,
    description: str = ...,
    long_description: str = ...,
    long_description_content_type: str = ...,
    author: str = ...,
    author_email: str = ...,
    maintainer: str = ...,
    maintainer_email: str = ...,
    url: str = ...,
    download_url: str = ...,
    packages: list[str] = ...,
    py_modules: list[str] = ...,
    scripts: list[str] = ...,
    ext_modules: Sequence[Extension] = ...,
    classifiers: list[str] = ...,
    distclass: type[Distribution] = ...,
    script_name: str = ...,
    script_args: list[str] = ...,
    options: Mapping[str, Any] = ...,
    license: str = ...,
    keywords: list[str] | str = ...,
    platforms: list[str] | str = ...,
    cmdclass: Mapping[str, type[_Command]] = ...,
    data_files: list[tuple[str, list[str]]] = ...,
    package_dir: Mapping[str, str] = ...,
    obsoletes: list[str] = ...,
    provides: list[str] = ...,
    requires: list[str] = ...,
    command_packages: list[str] = ...,
    command_options: Mapping[str, Mapping[str, tuple[Any, Any]]] = ...,
    package_data: Mapping[str, list[str]] = ...,
    include_package_data: bool = ...,
    libraries: list[str] = ...,
    headers: list[str] = ...,
    ext_package: str = ...,
    include_dirs: list[str] = ...,
    password: str = ...,
    fullname: str = ...,
    **attrs: Any,
) -> Distribution: ...

class Command(_Command):
    command_consumes_arguments: bool
    distribution: Distribution
    def __init__(self, dist: Distribution, **kw: Any) -> None: ...
    def ensure_string_list(self, option: str | list[str]) -> None: ...
    def reinitialize_command(
        self, command: _Command | str, reinit_subcommands: bool | Literal[0, 1] = 0, **kw: Any
    ) -> _Command: ...
    @abstractmethod
    def initialize_options(self) -> None: ...
    @abstractmethod
    def finalize_options(self) -> None: ...
    @abstractmethod
    def run(self) -> None: ...


from typehintwhl.command.build_stub import build_stub

import setuptools

__all__ = ["__version__", "setup"]

__version__ = "0.0.0"

def setup(**attrs):
    cmdclass = attrs.setdefault("cmdclass", {})
    cmdclass.setdefault("build_stub", build_stub)
    return setuptools.setup(**attrs)